# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from decimal import Decimal

from .Constants import DaxMethodIds, DaxDataRequestParam
from .DaxError import DaxClientError, DaxErrorCode

NAME_PREFIX = '#key'
VALUE_PREFIX = ':val'


def convert_request(request, method_id):
    # Make a copy so we don't modify the user's instance
    request = request.copy()

    # Workaround a bug in the initial release of DAX
    request.setdefault('ReturnConsumedCapacity', 'NONE')

    convert_V1_to_V2_request(request, method_id)

    return request


def is_V1_request(request, method_id):
    if method_id == DaxMethodIds.getItem_263244906_1_Id:
        return 'AttributesToGet' in request
    elif method_id == DaxMethodIds.batchGetItem_N697851100_1_Id:
        items = request.get('RequestItems')
        if items:
            for _, table_req in items.items():
                if 'AttributesToGet' in table_req:
                    return True
        return False
    elif method_id in (DaxMethodIds.putItem_N2106490455_1_Id, DaxMethodIds.deleteItem_1013539361_1_Id):
        return 'Expected' in request
    elif method_id == DaxMethodIds.updateItem_1425579023_1_Id:
        return 'Expected' in request or 'AttributeUpdates' in request
    elif method_id == DaxMethodIds.query_N931250863_1_Id:
        return 'AttributesToGet' in request or 'QueryFilter' in request or 'KeyConditions' in request
    elif method_id == DaxMethodIds.scan_N1875390620_1_Id:
        return 'AttributesToGet' in request or 'ScanFilter' in request
    elif method_id == DaxMethodIds.batchWriteItem_116217951_1_Id:
        return False
    else:
        raise DaxClientError('Unknown method_id type {}'.format(method_id), DaxErrorCode.Validation, False)


def convert_V1_to_V2_request(request, method_id):
    if method_id == DaxMethodIds.getItem_263244906_1_Id:
        attr_to_get = request.pop('AttributesToGet', None)
        if attr_to_get:
            _convert_AttributesToGet_to_ProjectionExpression_request(request, attr_to_get)

    elif method_id == DaxMethodIds.batchGetItem_N697851100_1_Id:
        items = request.get('RequestItems')
        if items:
            for table_name, table_req in items.items():
                attr_to_get = table_req.pop('AttributesToGet', None)
                if attr_to_get:
                    _convert_AttributesToGet_to_ProjectionExpression_request(table_req, attr_to_get)

    elif method_id in (DaxMethodIds.putItem_N2106490455_1_Id, DaxMethodIds.deleteItem_1013539361_1_Id):
        expected = request.pop('Expected', None)
        if expected:
            _convert_Expected_to_ConditionExpression_request(request, expected)

    elif method_id == DaxMethodIds.updateItem_1425579023_1_Id:
        expected = request.pop('Expected', None)
        if expected:
            _convert_Expected_to_ConditionExpression_request(request, expected)

        attr_updates = request.pop('AttributeUpdates', None)
        if attr_updates:
            _convert_AttributeUpdates_to_UpdateExpression_request(request, attr_updates)

    elif method_id == DaxMethodIds.query_N931250863_1_Id:

        attr_to_get = request.pop('AttributesToGet', None)
        if attr_to_get:
            _convert_AttributesToGet_to_ProjectionExpression_request(request, attr_to_get)

        key_conds = request.pop('KeyConditions', None)
        if key_conds:
            _convert_KeyConditions_to_Expression(request, key_conds)

        query_filter = request.pop('QueryFilter', None)
        if query_filter:
            _convert_Filter_to_Expression(request, query_filter)

    elif method_id == DaxMethodIds.scan_N1875390620_1_Id:
        attr_to_get = request.pop('AttributesToGet', None)
        if attr_to_get:
            _convert_AttributesToGet_to_ProjectionExpression_request(request, attr_to_get)

        scan_filter = request.pop('ScanFilter', None)
        if scan_filter:
            _convert_Filter_to_Expression(request, scan_filter)

    elif method_id == DaxMethodIds.batchWriteItem_116217951_1_Id:
        pass
    elif method_id == DaxMethodIds.transactWriteItems_N1160037738_1_Id:
        pass
    elif method_id == DaxMethodIds.transactGetItems_1866287579_1_Id:
        pass
    else:
        raise DaxClientError('Unknown method_id type {}'.format(method_id), DaxErrorCode.Validation, False)

    # Clean up remaining V1 attributes
    request.pop('ConditionalOperator', None)


def _convert_AttributesToGet_to_ProjectionExpression_request(request, attr_to_get):
    expr_attr_names = request.setdefault('ExpressionAttributeNames', {})
    request['ProjectionExpression'] = _convert_to_expr_attr_string(attr_to_get, NAME_PREFIX, ',', expr_attr_names)


def _convert_Expected_to_ConditionExpression_request(request, expected):
    expr_attr_names = request.setdefault('ExpressionAttributeNames', {})
    expr_attr_values = request.setdefault('ExpressionAttributeValues', {})
    cond_op = request.get('ConditionalOperator', None)
    request['ConditionExpression'] = _convert_conditions_to_expr(
        expected,
        cond_op,
        expr_attr_names,
        expr_attr_values,
        DaxDataRequestParam.ConditionExpression)


def _convert_AttributeUpdates_to_UpdateExpression_request(request, attr_updates):
    expr_attr_names = request.setdefault('ExpressionAttributeNames', {})
    expr_attr_values = request.setdefault('ExpressionAttributeValues', {})
    request['UpdateExpression'] = _convert_update_attr_to_update_expr(attr_updates, expr_attr_names, expr_attr_values)


def _convert_KeyConditions_to_Expression(request, key_conds):
    expr_attr_names = request.setdefault('ExpressionAttributeNames', {})
    expr_attr_values = request.setdefault('ExpressionAttributeValues', {})
    cond_op = request.get('ConditionalOperator', None)
    request['KeyConditionExpression'] = _convert_conditions_to_expr(
        key_conds,
        cond_op,
        expr_attr_names,
        expr_attr_values,
        DaxDataRequestParam.KeyConditionExpression)


def _convert_Filter_to_Expression(request, filter_):
    expr_attr_names = request.setdefault('ExpressionAttributeNames', {})
    expr_attr_values = request.setdefault('ExpressionAttributeValues', {})
    cond_op = request.get('ConditionalOperator', None)
    request['FilterExpression'] = _convert_conditions_to_expr(
        filter_,
        cond_op,
        expr_attr_names,
        expr_attr_values,
        DaxDataRequestParam.FilterExpression)


def _convert_update_attr_to_update_expr(attr_updates, expr_attr_names, expr_attr_values):
    sets = []
    adds = []
    deletes = []
    removes = []

    for attr_name, attr_update in sorted(attr_updates.items()):
        # Move ahead If update entry is null. DynamoDB's behavior.
        if not attr_update:
            continue

        action = attr_update.get('Action', 'PUT')
        attr_value = attr_update.get('Value')
        if not attr_value and action != 'DELETE':
            raise DaxClientError(
                'Only DELETE action is allowed when no attribute value is specified: ' + attr_name,
                DaxErrorCode.Validation, False)

        if action == 'PUT':
            put_expr = '{} = {}'.format(
                _append_to_expr_attr_map(attr_name, expr_attr_names, NAME_PREFIX),
                _append_to_expr_attr_map(attr_value, expr_attr_values, VALUE_PREFIX))
            sets.append(put_expr)
        elif action == 'ADD':
            add_expr = '{} {}'.format(
                _append_to_expr_attr_map(attr_name, expr_attr_names, NAME_PREFIX),
                _append_to_expr_attr_map(attr_value, expr_attr_values, VALUE_PREFIX))
            adds.append(add_expr)
        elif action == 'DELETE':
            if attr_value:
                del_expr = '{} {}'.format(
                    _append_to_expr_attr_map(attr_name, expr_attr_names, NAME_PREFIX),
                    _append_to_expr_attr_map(attr_value, expr_attr_values, VALUE_PREFIX))
                deletes.append(del_expr)
            else:
                rem_expr = _append_to_expr_attr_map(attr_name, expr_attr_names, NAME_PREFIX)
                removes.append(rem_expr)
        else:
            raise DaxClientError(
                'Action must be one of (ADD, DELETE, PUT) for {}, got {}'.format(attr_name, action),
                DaxErrorCode.Validation, False)

    update_expr = ' '.join((
        _generate_and_append_update_expr(sets, 'SET'),
        _generate_and_append_update_expr(adds, 'ADD'),
        _generate_and_append_update_expr(deletes, 'DELETE'),
        _generate_and_append_update_expr(removes, 'REMOVE')))

    return update_expr


def _generate_and_append_update_expr(items, op):
    return op + ' ' + ', '.join(items) if items else ''


def _convert_conditions_to_expr(conditions, operator, expr_attr_names, expr_attr_values, expr_type):
    operator = operator or 'AND'
    operator = ' ' + operator.strip() + ' '

    if expr_type == DaxDataRequestParam.FilterExpression:
        encoder = _encode_filter_expr
    elif expr_type == DaxDataRequestParam.ConditionExpression:
        encoder = _encode_condition_expr
    elif expr_type == DaxDataRequestParam.KeyConditionExpression:
        encoder = _encode_key_condition_expr
    else:
        raise ValueError('Unknown expression type: ' + expr_type)

    # Join all non-empty conditions
    return operator.join(filter(None,
                                (encoder(cond_attr, cond_details, expr_attr_names, expr_attr_values) \
                                 for cond_attr, cond_details in sorted(conditions.items()))
                                ))


def _encode_filter_expr(attr, condition, expr_attr_names, expr_attr_values):
    if condition is None:
        # This is a special case that is filtered by other clients, so remove it as well
        return

    name = _append_to_expr_attr_map(attr, expr_attr_names, NAME_PREFIX)
    comp_op = condition.get('ComparisonOperator')
    attr_value_list = condition.get('AttributeValueList')
    if attr_value_list and not comp_op:
        raise DaxClientError(
            'One or more parameter values were invalid: AttributeValueList can only be used with a ComparisonOperator for Attribute: ' + attr,
            DaxErrorCode.Validation, False)

    return _construct_filter_or_expected_comp_op_expr(
        comp_op, name, attr, attr_value_list, expr_attr_values, 'Unsupported operator on ExpectedAttributeValue: ' + str(condition))


def _encode_condition_expr(attr, expected_attr_val, expr_attr_names, expr_attr_values):
    name = _append_to_expr_attr_map(attr, expr_attr_names, NAME_PREFIX)
    comp_op = expected_attr_val.get('ComparisonOperator')
    if not comp_op:
        return _handle_exists_criteria(attr, expected_attr_val, name, expr_attr_values)

    if 'Exists' in expected_attr_val:
        raise DaxClientError(
            'One or more parameter values were invalid: Exists and ComparisonOperator cannot be used together for Attribute: ' + attr,
            DaxErrorCode.Validation, False)

    attr_value_list = expected_attr_val.get('AttributeValueList')
    if not attr_value_list:
        value = expected_attr_val.get('Value')
        if value:
            attr_value_list = [value]
    else:
        if 'Value' in expected_attr_val:
            raise DaxClientError(
                'Value and AttributeValueList cannot be used together for Attribute: {} {}'.format(attr, attr_value_list),
                DaxErrorCode.Validation, False)

    return _construct_filter_or_expected_comp_op_expr(comp_op, name, attr, attr_value_list, expr_attr_values,
                                                      'Unsupported operator on ExpectedAttributeValue: ' + str(expected_attr_val))


def _encode_key_condition_expr(attr, key_cond, expr_attr_names, expr_attr_values):
    if not key_cond:
        raise DaxClientError('KeyCondition cannot be null for key: ' + attr, DaxErrorCode.Validation, False)

    comp_op = key_cond.get('ComparisonOperator')
    if not comp_op:
        raise DaxClientError('ComparisonOperator cannot be empty for KeyCondition: ' + key_cond, DaxErrorCode.Validation, False)

    attr_value_list = key_cond['AttributeValueList']
    name = _append_to_expr_attr_map(attr, expr_attr_names, NAME_PREFIX)

    if comp_op == 'BETWEEN':
        return _handle_between_condition(comp_op, name, attr_value_list, expr_attr_values, attr)
    elif comp_op == 'BEGINS_WITH':
        return _handle_begins_with_condition(comp_op, name, attr_value_list, expr_attr_values, attr)
    elif comp_op == 'EQ':
        op = '='
    elif comp_op == 'LE':
        op = '<='
    elif comp_op == 'LT':
        op = '<'
    elif comp_op == 'GE':
        op = '>='
    elif comp_op == 'GT':
        op = '>'
    else:
        raise DaxClientError('Unsupported operator on KeyCondition: ' + comp_op, DaxErrorCode.Validation, False)

    _check_num_arguments(comp_op, 1, attr_value_list, attr)
    av0 = attr_value_list[0]
    expr_attr_val0 = _append_to_expr_attr_map(av0, expr_attr_values, VALUE_PREFIX)
    return name + ' ' + op + ' ' + expr_attr_val0


def _construct_filter_or_expected_comp_op_expr(comp_op, name, attr, attr_value_list, expr_attr_values, error_msg):
    if comp_op == 'BETWEEN':
        return _handle_between_condition(comp_op, name, attr_value_list, expr_attr_values, attr)
    elif comp_op == 'BEGINS_WITH':
        return _handle_begins_with_condition(comp_op, name, attr_value_list, expr_attr_values, attr)
    elif comp_op == 'NOT_CONTAINS':  # Don't append key name
        return 'attribute_exists(' + name + ') AND NOT ' + _handle_contains_condition(comp_op, name, attr_value_list, expr_attr_values,
                                                                                      attr)
    elif comp_op == 'CONTAINS':  # Don't append key name
        return _handle_contains_condition(comp_op, name, attr_value_list, expr_attr_values, attr)
    elif comp_op == 'NOT_NULL':  # Don't append key name
        _check_num_arguments(comp_op, 0, attr_value_list, attr)
        return 'attribute_exists(' + name + ')'
    elif comp_op == 'NULL':  # Don't append key name
        _check_num_arguments(comp_op, 0, attr_value_list, attr)
        return 'attribute_not_exists(' + name + ')'
    elif comp_op == 'IN':
        return _handle_in_condition(comp_op, name, attr_value_list, expr_attr_values, attr)
    elif comp_op == 'EQ':
        op = '='
    elif comp_op == 'LE':
        op = '<='
    elif comp_op == 'LT':
        op = '<'
    elif comp_op == 'GE':
        op = '>='
    elif comp_op == 'GT':
        op = '>'
    elif comp_op == 'NE':
        op = '<>'
    else:
        raise DaxClientError(error_msg, DaxErrorCode.Validation, False)

    _check_num_arguments(comp_op, 1, attr_value_list, attr)
    av0 = attr_value_list[0]
    _check_valid_BSN_type(comp_op, av0)
    expr_attr_val0 = _append_to_expr_attr_map(av0, expr_attr_values, VALUE_PREFIX)
    return '{} {} {}'.format(name, op, expr_attr_val0)


def _convert_to_expr_attr_string(attrs, prefix, delim, expr_attr_names):
    return delim.join(_append_to_expr_attr_map(attr, expr_attr_names, prefix) for attr in attrs)


def _append_to_expr_attr_map(attr, expr_attr_names, prefix):
    suffix = len(expr_attr_names)
    name = prefix + str(suffix)
    while name in expr_attr_names:
        # if it already exists, find a new one
        suffix += 1
        name = prefix + str(suffix)

    expr_attr_names[name] = attr
    return name


def _handle_exists_criteria(attr, expected_attr_val, name, expr_attr_values):
    attr_val_list = expected_attr_val.get('AttributeValueList')
    if attr_val_list:
        raise DaxClientError(
            'One or more parameter values were invalid: AttributeValueList can only be used with a ComparisonOperator for Attribute: ' + attr,
            DaxErrorCode.Validation, False)

    exists = expected_attr_val.get('Exists')  # Exists defaults to true if not specified
    value = expected_attr_val.get('Value')
    if exists is None or exists:
        # If Exists is true, a value must be provided
        if exists and not value:
            raise DaxClientError(
                'One or more parameter values were invalid: Value must be provided when Exists is true for Attribute: ' + attr,
                DaxErrorCode.Validation, False)

        if exists is None and not value:
            raise DaxClientError(
                'One or more parameter values were invalid: Value must be provided when Exists is null for Attribute: ' + attr,
                DaxErrorCode.Validation, False)

        expected_attr_val0 = _append_to_expr_attr_map(value, expr_attr_values, VALUE_PREFIX)
        return name + ' = ' + expected_attr_val0
    else:
        # If Exists is false they must not provide a value
        if value:
            raise DaxClientError(
                'One or more parameter values were invalid: Value cannot be used when Exists is false for Attribute: ' + attr,
                DaxErrorCode.Validation, False)

        return 'attribute_not_exists(' + name + ')'


def _handle_between_condition(op, name, attr_value_list, expr_attr_values, attr):
    _check_num_arguments(op, 2, attr_value_list, attr)
    _check_valid_BSN_type(op, attr_value_list[0], attr_value_list[1])
    _check_valid_bounds(attr_value_list[0], attr_value_list[1])
    expr_attr_val0 = _append_to_expr_attr_map(attr_value_list[0], expr_attr_values, VALUE_PREFIX)
    expr_attr_val1 = _append_to_expr_attr_map(attr_value_list[1], expr_attr_values, VALUE_PREFIX)
    return '{} between {} AND {}'.format(name, expr_attr_val0, expr_attr_val1)


def _handle_begins_with_condition(op, name, attr_value_list, expr_attr_values, attr):
    _check_num_arguments(op, 1, attr_value_list, attr)
    av0 = attr_value_list[0]
    if 'B' not in av0 and 'S' not in av0:
        raise DaxClientError(
            'One or more parameter values were invalid: ComparisonOperator {} is not valid for {} AttributeValue type'.format(
                op, list(av0.keys())[0]),
            DaxErrorCode.Validation, False)

    expr_attr_val0 = _append_to_expr_attr_map(av0, expr_attr_values, VALUE_PREFIX)
    return 'begins_with({}, {})'.format(name, expr_attr_val0)


def _handle_contains_condition(op, name, attr_value_list, expr_attr_values, attr):
    _check_num_arguments(op, 1, attr_value_list, attr)
    av0 = attr_value_list[0]
    _check_valid_BSNBoolNull_types(op, av0)
    expr_attr_val0 = _append_to_expr_attr_map(av0, expr_attr_values, VALUE_PREFIX)
    return 'contains({}, {})'.format(name, expr_attr_val0)


def _handle_in_condition(op, name, attr_value_list, expr_attr_values, attr):
    if attr_value_list is None:
        raise DaxClientError(
            'One or more parameter values were invalid: Value or AttributeValueList must be used with ComparisonOperator: IN for Attribute: ' + attr,
            DaxErrorCode.Validation, False)
    else:
        if len(attr_value_list) == 0:
            raise DaxClientError(
                'One or more parameter values were invalid: Invalid number of argument(s) for the IN ComparisonOperator',
                DaxErrorCode.Validation, False)

    _check_valid_BSN_type(op, *attr_value_list)
    return '{} IN ({})'.format(name, _convert_to_expr_attr_string(attr_value_list, VALUE_PREFIX, ',', expr_attr_values))


def _check_num_arguments(op, expected_args_count, attr_value_list, attr):
    size = len(attr_value_list) if attr_value_list else 0
    if not attr_value_list and expected_args_count > 0:
        raise DaxClientError(
            'One or more parameter values were invalid: Value or AttributeValueList must be used with ComparisonOperator: ' + op + ' for Attribute: ' + attr,
            DaxErrorCode.Validation, False)

    if size != expected_args_count:
        raise DaxClientError(
            'One or more parameter values were invalid: Invalid number of argument(s) for the ' + op + ' ComparisonOperator',
            DaxErrorCode.Validation, False)


def _check_valid_BSN_type(op, *attr_vals):
    if op in ('NE', 'EQ'):
        return

    for attr_val in attr_vals:
        if 'B' not in attr_val and 'S' not in attr_val and 'N' not in attr_val:
            raise DaxClientError(
                'One or more parameter values were invalid: ComparisonOperator ' + op + ' is not valid for ' + str(
                    list(attr_val.keys())) + ' AttributeValue type',
                DaxErrorCode.Validation, False)


def _check_valid_BSNBoolNull_types(op, *attr_vals):
    for attr_val in attr_vals:
        if 'BOOL' in attr_val or 'NULL' in attr_val:
            continue
        _check_valid_BSN_type(op, attr_val)


def _check_valid_bounds(lower_av, upper_av):
    lb_type = list(lower_av.keys())[0]
    ub_type = list(upper_av.keys())[0]

    if lb_type != ub_type:
        raise DaxClientError(
            'One or more parameter values were invalid: AttributeValues inside AttributeValueList must be of same type',
            DaxErrorCode.Validation, False)

    if lb_type == 'S':
        if lower_av['S'] > upper_av['S']:
            raise DaxClientError(
                'The BETWEEN condition was provided a range where the lower bound is greater than the upper bound',
                DaxErrorCode.Validation, False)
    elif lb_type == 'N':
        if Decimal(lower_av['N']) > Decimal(upper_av['N']):
            raise DaxClientError(
                'The BETWEEN condition was provided a range where the lower bound is greater than the upper bound',
                DaxErrorCode.Validation, False)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from collections import defaultdict

from botocore.exceptions import ClientError as BotoClientError


class DaxErrorCode(object):
    Decoder = 'DecoderException'
    Unrecognized = 'UnrecognizedClientException'
    Authentication = 'MissingAuthenticationTokenException'
    MalformedResult = 'MalformedResultException'
    EndOfStream = 'EndOfStreamException'
    IllegalArgument = 'IllegalArgumentException'
    Validation = 'ValidationException'
    NoRoute = 'NoRouteException'
    ResourceNotFound = 'ResourceNotFoundException'
    ResourceInUse = 'ResourceInUseException'
    ProvisionedThroughputExceeded = 'ProvisionedThroughputExceededException'
    ConditionalCheckFailed = 'ConditionalCheckFailedException'
    InternalServerError = 'InternalServerErrorException'
    ItemCollectionSizeLimitExceeded = 'ItemCollectionSizeLimitExceededException'
    LimitExceeded = 'LimitExceededException'
    Throttling = 'ThrottlingException'
    AccessDenied = 'AccessDeniedException'
    TransactionCanceledException = 'TransactionCanceledException'
    RequestLimitExceededException = 'RequestLimitExceededException'


class DaxClientError(BotoClientError):
    # `retryable` is not used, but removing it causes the integration tests to fail, TODO fix
    def __init__(self, message, ddb_code, retryable=False): # pylint: disable=unused-argument
        error_response, code = _make_error_response(
            message, None, None, ddb_code, None)
        super(DaxClientError, self).__init__(error_response, '')

        self.code = code
        self.codes = None
        self.http_status = None
        self.auth_error = False
        self.wait_for_recovery_before_retry = False


class DaxValidationError(DaxClientError):
    def __init__(self, message):
        super(DaxValidationError, self).__init__(
            message, DaxErrorCode.Validation, False)


class DaxServiceError(BotoClientError):
    def __init__(self, operation_name, message, dax_codes, request_id, ddb_code, http_status, cancellation_reasons=None):
        error_response, code = _make_error_response(
            message, dax_codes, request_id, ddb_code, http_status)
        super(DaxServiceError, self).__init__(error_response, operation_name)

        self.code = code
        self.codes = dax_codes
        self.http_status = http_status
        self.cancellation_reasons = cancellation_reasons
        self.auth_error = _determine_auth_error(self.codes)
        self.wait_for_recovery_before_retry = _determine_wait_for_recovery_before_retry(
            self.codes)

    @classmethod
    def from_no_more_data_exception(cls, operation_name, exc):
        '''Return an exception wrapping a no more data exception.

        Necessary to avoid cluttering up the DaxServiceError constructor, which
        assumes the presence of error codes from DAX.

        '''

        dse = DaxServiceError(
            operation_name, "remote socket is closed", (), None, None, None)
        dse.__cause__ = exc
        dse.wait_for_recovery_before_retry = True
        return dse


def _make_error_response(message, codes, request_id, ddb_code, http_status):
    # Make an error response that mimics Boto's for compatibility
    def yodict(): return defaultdict(yodict)  # A recursive default dict
    response = yodict()

    code = _pick_error_code(codes) or ddb_code
    if code is not None:
        response['Error']['Code'] = code

    if message:
        response['Error']['Message'] = message

    if request_id:
        response['ResponseMetadata']['RequestId'] = request_id

    if http_status:
        response['ResponseMetadata']['HTTPStatusCode'] = http_status

    # Convert to a regular dict for compatibility
    return _unwrap_defaultdict(response), code


def _unwrap_defaultdict(d):
    if isinstance(d, defaultdict):
        return {k: _unwrap_defaultdict(v) for k, v in d.items()}
    else:
        return d


def _pick_error_code(codes):
    if not codes:
        return None

    if codes == [3, 37, 54]:
        return DaxErrorCode.InternalServerError

    if codes[0] != 4:
        return None

    code = None
    lookup = ERROR_MAP
    for e in codes[1:]:
        lookup = lookup.get(e)
        if lookup is None:
            break
    else:
        code = lookup

    return code


def _determine_auth_error(codes):
    return codes is not None \
        and len(codes) >= 3 \
        and codes[1] == 23 and codes[2] == 31


def _determine_wait_for_recovery_before_retry(codes):
    return len(codes) >= 1 and codes[0] == 2


def is_retryable(exc):
    """
        The Dax client will retry event if it is a retryable exception.
        This function is modeled after the functionality in the Java Client: https://github.com/aws/aws-sdk-java/blob/master/aws-java-sdk-core/src/main/java/com/amazonaws/retry/RetryUtils.java#L77
        Right now, it will retry most error codes.
        Codes arrays whose first value is a "3" are classified as an "Unretryable server error." according to DaxCore.
    """
    if not isinstance(exc, (DaxClientError, DaxServiceError)):
        return False

    return (exc.codes and exc.codes[0] != 3) \
        or exc.code and exc.code in RETRYABLE_ERRORS \
        or exc.http_status in (429, 500, 503)


def is_retryable_with_backoff(exc):
    """
        The Dax client will backoff and jitter a retry event if it is a throttling exception.
        This function is modeled after the functionality in the Java Client: https://github.com/aws/aws-sdk-java/blob/master/aws-java-sdk-core/src/main/java/com/amazonaws/retry/RetryUtils.java#L98
    """
    if not isinstance(exc, (DaxClientError, DaxServiceError)):
        return False

    return exc.code in RETRYABLE_ERRORS_WITH_THROTTLE \
        or exc.http_status == 429


ERROR_MAP = {
    23: {
        24: DaxErrorCode.ResourceNotFound,
        35: DaxErrorCode.ResourceInUse,
    },
    37: {
        38: {
            39: {
                40: DaxErrorCode.ProvisionedThroughputExceeded,
                41: DaxErrorCode.ResourceNotFound,
                43: DaxErrorCode.ConditionalCheckFailed,
                45: DaxErrorCode.ResourceInUse,
                46: DaxErrorCode.Validation,
                47: DaxErrorCode.InternalServerError,
                48: DaxErrorCode.ItemCollectionSizeLimitExceeded,
                49: DaxErrorCode.LimitExceeded,
                50: DaxErrorCode.Throttling,
                58: DaxErrorCode.TransactionCanceledException,
                61: DaxErrorCode.RequestLimitExceededException,
            },
            42: DaxErrorCode.AccessDenied,
            44: (DaxErrorCode.Validation, 'NotImplementedException'),
        }
    }
}

"""
    This set is modeled after the set exposed in the AWS Java SDK:
    https://github.com/aws/aws-sdk-java/blob/master/aws-java-sdk-core/src/main/java/com/amazonaws/retry/RetryUtils.java#L55
"""
RETRYABLE_ERRORS = {
    DaxErrorCode.InternalServerError,
    DaxErrorCode.TransactionCanceledException
}

"""
    This set is modeled after the set exposed in the AWS Java SDK:
    https://github.com/aws/aws-sdk-java/blob/master/aws-java-sdk-core/src/main/java/com/amazonaws/retry/RetryUtils.java#L31
"""
RETRYABLE_ERRORS_WITH_THROTTLE = {
    DaxErrorCode.Throttling,
    DaxErrorCode.ProvisionedThroughputExceeded,
    DaxErrorCode.RequestLimitExceededException,
    DaxErrorCode.LimitExceeded,
}



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from __future__ import print_function

import socket

from collections import defaultdict

from . import AttributeValueDecoder
from .Constants import DaxResponseParam, CONSUMED_CAPACITY
from .DaxError import DaxValidationError, DaxErrorCode

# Do not re-order these fields; the indexes match the values hard-coded on the server
ENDPOINT_FIELDS = ('node', 'hostname', 'address', 'port', 'role', 'az', 'leader_session_id')
_ADDRESS_FIELD_IX = ENDPOINT_FIELDS.index('address')
_IP_ADDRESS_FMT = '{:d}.{:d}.{:d}.{:d}'


def endpoints_455855874_1(_, tube):
    endpoints = tube.read_array()
    return [{ENDPOINT_FIELDS[k]: _fixup_endpoint(k, v) for k, v in ep.items()} for ep in endpoints]


def _fixup_endpoint(key, value):
    if key == _ADDRESS_FIELD_IX:
        # Return the address as a string, as that is what Python's connect method uses
        return socket.inet_ntoa(bytes(value))
    else:
        return value


def defineKeySchema_N742646399_1(_, tube):
    schemaMap = tube.read_map()
    return [{'AttributeName': attrName, 'AttributeType': attrType} for attrName, attrType in schemaMap.items()]


def defineAttributeList_670678385_1(_, tube):
    return tube.read_array()


def defineAttributeListId_N1230579644_1(_, tube):
    return tube.read_int()


def batchGetItem_N697851100_1(request, tube):
    size = tube.dec._decode_array_header()
    if size != 2:
        raise DaxValidationError("Incorrect array size: " + str(size))

    # result[0] -> Responses
    responses = {}
    for table_dec in tube.dec.decode_map_iter():
        table_name = table_dec.decode_string()
        proj_ordinals = request._projection_ordinals_by_table.get(table_name)
        items = []

        if proj_ordinals:
            for item_dec in table_dec.decode_array_iter():
                item = AttributeValueDecoder._decode_item_internal(item_dec, request, table_name)
                items.append(item)
        else:
            key_schema = request.key_schema_by_table[table_name]
            num_items = table_dec._decode_array_header()
            for _ in range(num_items, 0, -2):
                key = AttributeValueDecoder._decode_key_bytes(table_dec, key_schema)
                item = AttributeValueDecoder._decode_stream_item(table_dec.decode_cbor())
                item.update(key)
                items.append(item)

        responses[table_name] = items

    # result[1] -> UnprocessedKeys
    unprocessed_keys = {}
    for upk_dec in tube.dec.decode_map_iter():
        request_info = {}
        table_name = upk_dec.decode_string()
        key_schema = request.key_schema_by_table[table_name]
        keys = [AttributeValueDecoder._decode_key_bytes(key_dec, key_schema) \
                for key_dec in upk_dec.decode_array_iter()]

        if keys:
            request_info['Keys'] = keys

            for request_field in ('ProjectionExpression', 'ConsistentRead', 'AttributesToGet', 'ExpressionAttributeNames'):
                if request_field in request.RequestItems[table_name] and request.RequestItems[table_name][request_field]:
                    request_info[request_field] = request.RequestItems[table_name][request_field]

            if request_info:
                unprocessed_keys[table_name] = request_info

    result = {
        'Responses': responses,
        'UnprocessedKeys': unprocessed_keys
    }

    # ConsumedCapacity
    consumed_capacity = [_decode_consumed_capacity(cc_dec.decode_cbor()) for cc_dec in tube.dec.decode_array_iter()]
    if 'ReturnConsumedCapacity' in request and request.ReturnConsumedCapacity != 'NONE':
        # Handle cases where no CC is returned b/c table came from cache
        missing_tables = set(request.RequestItems.keys()) - set(cc['TableName'] for cc in consumed_capacity)
        consumed_capacity.extend({
                                     "TableName": table_name,
                                     "CapacityUnits": 0.0
                                 } for table_name in missing_tables)

        result['ConsumedCapacity'] = consumed_capacity

    return result


def transactGetItems_1866287579_1(request, tube):
    result = {}

    response_length = tube.dec._decode_array_header()
    if response_length != 2:
        raise DaxValidationError(
            "TransactGetResponse needs to have 2 elements, instead had {}. {}".format(response_length, DaxErrorCode.MalformedResult))

    # decode responses
    responses = []
    i = 0
    for item_dec in tube.dec.decode_array_iter():
        if item_dec.try_decode_null():
            responses.append({})
        else:
            item = None
            request_data = request['TransactItems'][i]['Get']

            if 'projection_ordinals' in request_data:
                item = AttributeValueDecoder._decode_projection(item_dec, request_data['projection_ordinals'])
            else:
                item = AttributeValueDecoder._decode_stream_item(item_dec.decode_cbor())
                item.update(request_data['Key'])

            responses.append({'Item': item})
        i += 1
    result['Responses'] = responses

    # decode consumed capacity
    decoded_consumed_capacity = tube.dec.decode_array()
    consumed_capacity = [_decode_consumed_capacity_from_dict(_cc) for _cc in decoded_consumed_capacity]
    if consumed_capacity:
        result['ConsumedCapacity'] = consumed_capacity

    return result


def transactWriteItems_N1160037738_1(request, tube):
    result = {}
    response_length = tube.dec._decode_array_header()
    if response_length != 3:
        raise DaxValidationError(
            "TransactWriteResponse needs to have 3 elements, instead had {}. {}".format(response_length, DaxErrorCode.MalformedResult))

    # TODO: The first element in array is always item responses but we don't receive these for now. Hence skip the first element
    tube.dec.skip()

    # decode consumed capacity
    decoded_consumed_capacity = tube.dec.decode_array()
    consumed_capacity = [_decode_consumed_capacity_from_dict(_cc) for _cc in decoded_consumed_capacity]
    if consumed_capacity:
        result['ConsumedCapacity'] = consumed_capacity

    # decode item collection metrics
    item_collection_metrics = defaultdict(list)
    for table_dec in tube.dec.decode_map_iter():
        table_name = table_dec.decode_string()
        key_schema = request.key_schema_by_table[table_name]
        for ic_dec in table_dec.decode_array_iter():
            item_collection_metrics[table_name].append(_decode_item_collection_metrics(ic_dec, key_schema))

    if item_collection_metrics:
        result['ItemCollectionMetrics'] = dict(item_collection_metrics)

    return result


def batchWriteItem_116217951_1(request, tube):
    unprocessed_items_by_table = defaultdict(list)

    for _dec in tube.dec.decode_map_iter():
        table_name = _dec.decode_string()
        num_items = _dec._decode_array_header()
        key_schema = request.key_schema_by_table[table_name]

        for _ in range(num_items, 0, -2):
            key = AttributeValueDecoder._decode_key_bytes(_dec, key_schema)
            if _dec.try_decode_null():
                # DeleteRequest
                unprocessed_items_by_table[table_name].append({'DeleteRequest': {'Key': key}})
            else:
                # PutRequest
                item = AttributeValueDecoder._decode_stream_item(_dec.decode_cbor())
                # These get de-anonymized later
                unprocessed_items_by_table[table_name].append({'PutRequest': {'Item': item}})

    result = {
        'UnprocessedItems': dict(unprocessed_items_by_table)
    }

    consumed_capacity = [_decode_consumed_capacity(cc_dec.decode_cbor()) for cc_dec in tube.dec.decode_array_iter()]

    if 'ReturnConsumedCapacity' in request and request.ReturnConsumedCapacity != 'NONE' and consumed_capacity:
        result['ConsumedCapacity'] = consumed_capacity

    item_collection_metrics = defaultdict(list)

    for table_dec in tube.dec.decode_map_iter():
        table_name = table_dec.decode_string()
        key_schema = request.key_schema_by_table[table_name]
        for ic_dec in table_dec.decode_array_iter():
            item_collection_metrics[table_name].append(_decode_item_collection_metrics(ic_dec, key_schema))

    if item_collection_metrics:
        result['ItemCollectionMetrics'] = item_collection_metrics

    return result


def getItem_263244906_1(request, tube):
    return _decodeNormalOperation(request, tube)


def putItem_N2106490455_1(request, tube):
    return _decodeNormalOperation(request, tube)


def updateItem_1425579023_1(request, tube):
    return _decodeNormalOperation(request, tube)


def deleteItem_1013539361_1(request, tube):
    return _decodeNormalOperation(request, tube)


def query_N931250863_1(request, tube):
    result = _decodeNormalOperation(request, tube)
    return result


def scan_N1875390620_1(request, tube):
    result = _decodeNormalOperation(request, tube)
    return result


def _decodeNormalOperation(request, tube):
    result = {}
    if tube.try_read_null():
        return result

    for _dec in tube.dec.decode_map_iter():
        param = _dec.decode_int()
        decoder = _PARAM_DECODERS[param]

        decoder(_dec, request, result)

    return result


def decode_Item(dec, request, result):
    result['Item'] = AttributeValueDecoder._decode_item_internal(dec, request)


def decode_Items(dec, request, result):
    result['Items'] = [AttributeValueDecoder._decode_item_internal(_dec, request) for _dec in dec.decode_array_iter()]


def decode_ConsumedCapacity(dec, request, result):
    if dec.try_decode_null():
        return

    consumed_capacity = _decode_consumed_capacity(dec.decode_cbor())
    if 'ReturnConsumedCapacity' in request and request.ReturnConsumedCapacity != 'NONE' and consumed_capacity:
        result['ConsumedCapacity'] = consumed_capacity


def _decode_consumed_capacity(dec):
    consumed_capacity = {}
    consumed_capacity['TableName'] = dec.decode_string()
    consumed_capacity['CapacityUnits'] = dec.decode_number()
    if not dec.try_decode_null():
        consumed_capacity['Table'] = {'CapacityUnits': dec.decode_number()}

    if not dec.try_decode_null():
        consumed_capacity['GlobalSecondaryIndexes'] = _decode_index_consumed_capacity(dec)

    if not dec.try_decode_null():
        consumed_capacity['LocalSecondaryIndexes'] = _decode_index_consumed_capacity(dec)

    return consumed_capacity


_INDEX_CODES = ('LocalSecondaryIndexes', 'GlobalSecondaryIndexes')


def _decode_consumed_capacity_from_dict(decoded_cc):
    consumed_capacity = {}
    decoded_cc = dict(decoded_cc)

    for cc_field, cc_code in CONSUMED_CAPACITY.items():
        if cc_code in decoded_cc:
            if cc_field == "Table":
                decoded_cc[cc_code] = _decode_consumed_capacity_from_dict(decoded_cc[cc_code])
            elif cc_field in _INDEX_CODES:
                decoded_cc[cc_code] = {k: _decode_consumed_capacity_from_dict(v) for k, v in decoded_cc[cc_code].items()}
            consumed_capacity[cc_field] = decoded_cc[cc_code]

    return consumed_capacity


def _decode_index_consumed_capacity(dec):
    idx_cc = {}
    for _dec in dec.decode_map_iter():
        index_name = _dec.decode_string()
        units = _dec.decode_number()
        idx_cc[index_name] = {'CapacityUnits': units}

    return idx_cc


def decode_Attributes(dec, request, result):
    return_values = request.get('ReturnValues')
    is_projection = return_values and return_values in ('UPDATED_NEW', 'UPDATED_OLD')

    if is_projection:
        item = AttributeValueDecoder._decode_stream_item_projection(dec.decode_cbor())
    else:
        item = AttributeValueDecoder._decode_stream_item(dec.decode_cbor())
        if item:
            AttributeValueDecoder._reinsert_key(item, request, request.key_schema)

    result['Attributes'] = item


def decode_Count(dec, _, result):
    result['Count'] = dec.decode_int()


def decode_ScannedCount(dec, _, result):
    result['ScannedCount'] = dec.decode_int()


def decode_LastEvaluatedKey(dec, request, result):
    if 'IndexName' in request:
        last_eval_key = AttributeValueDecoder._decode_compound_key(dec.decode_cbor())
    else:
        last_eval_key = AttributeValueDecoder._decode_key_bytes(dec, request.key_schema)

    result['LastEvaluatedKey'] = last_eval_key


def decode_ItemCollectionMetrics(dec, request, result):
    result['ItemCollectionMetrics'] = _decode_item_collection_metrics(dec, request.key_schema)


def _decode_item_collection_metrics(dec, key_schema):
    if dec.try_decode_null():
        return None

    _dec = dec.decode_cbor()
    key_av = AttributeValueDecoder._decode_attribute_value(_dec)
    size_lower = _dec.decode_float()
    size_upper = _dec.decode_float()
    item_collection_metrics = {
        'ItemCollectionKey': {key_schema[0]['AttributeName']: key_av},
        'SizeEstimateRangeGB': [size_lower, size_upper]
    }

    return item_collection_metrics


_PARAM_DECODERS = {
    DaxResponseParam.Item: decode_Item,
    DaxResponseParam.ConsumedCapacity: decode_ConsumedCapacity,
    DaxResponseParam.Attributes: decode_Attributes,
    DaxResponseParam.ItemCollectionMetrics: decode_ItemCollectionMetrics,
    DaxResponseParam.Items: decode_Items,
    DaxResponseParam.Count: decode_Count,
    DaxResponseParam.LastEvaluatedKey: decode_LastEvaluatedKey,
    DaxResponseParam.ScannedCount: decode_ScannedCount,
    # These are only in batch operations and are handled differently
    # DaxResponseParam.Responses: ...,
    # DaxResponseParam.UnprocessedKeys: ...,
    # DaxResponseParam.UnprocessedItems: ...,
}

_PARAM_NAME = {val: name for name, val in vars(DaxResponseParam).items() if not name.startswith('__')}



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import math
import struct
import six
from decimal import Decimal, localcontext

from . import compat
from .CborEncoder import DYNAMODB_CONTEXT

NULL_BYTE_LOW = 0       # Byte to use for undefined, low ordering
NULL_BYTE_HIGH = 0xff   # Byte to use for undefined, high ordering

def create_decimal(d):
    ''' Create Decimal instances using the proper context.

    This uses the same Context settings as botocore for consistency.
    '''
    return DYNAMODB_CONTEXT.create_decimal(d)

def encode(value, mask=0, context=None):
    if value is None:
        raise ValueError('value must not be None')

    if not isinstance(value, Decimal):
        value = create_decimal(value)

    if value == 0:
        return compat.bytelist([0x80 ^ mask])

    value_parts = value.as_tuple()
    encoded = bytearray(_estimateLength(value_parts))

    with localcontext(context or DYNAMODB_CONTEXT) as ctx:
        ctx.prec = ctx.prec + 2  # Include extra precision for handling unscaled values
        offset = _encode(value, value_parts, encoded, 0, mask)
    return bytes(encoded[:offset])

def _encode(value, value_parts, dst, offset, mask):
    precision = len(value_parts.digits)
    scale = -value_parts.exponent
    exponent = precision - scale

    # Do the header
    if value.is_signed():
        # Negative
        if exponent >= -0x3e and exponent < 0x3e:
            dst[offset] = (0x3f - exponent) ^ mask
            offset += 1
        else:
            dst[offset] = 0x7e ^ mask if exponent < 0 else 0x01 ^ mask
            # Force values to unsigned to write raw bits
            _writeInt32BE(dst, (exponent & 0xffffffff) ^ mask ^ 0x7fffffff, offset + 1)
            offset += 5
    else:
        # Positive or zero
        if exponent >= -0x3e and exponent < 0x3e:
            dst[offset] = (exponent + 0xc0) ^ mask
            offset += 1
        else:
            dst[offset] = 0x81 ^ mask if exponent < 0 else 0xfe ^ mask
            # Force values to unsigned to write raw bits
            _writeInt32BE(dst, (exponent & 0xffffffff) ^ mask ^ 0x80000000, offset + 1)
            offset += 5

    # Now do the significand.
    with localcontext() as ctx:
        # Use a higher scaling value for intermediate calcs
        ctx.Emax = max(ctx.Emax, scale)
        scaler = ctx.create_decimal((0, (1, ), scale))
        unscaled_value = value * scaler

    # Ensure a non-fractional amount of base-1000 digits.
    m = precision % 3
    if m == 1:
        terminator = 0
        unscaled_value *= 100
    elif m == 2:
        terminator = 1
        unscaled_value *= 10
    else:
        terminator = 2

    # 10 bits per base-1000 digit and 1 extra terminator digit. Digit values 0..999 are
    # encoded as 12..1011. Digit values 0..11 and 1012..1023 are used for terminators

    if not value.is_signed():
        # Positive
        digit_adj = 12
    else:
        digit_adj = 999 + 12
        terminator = 1023 - terminator

    pos = int(math.floor(math.ceil(precision / math.log(2) / 3.0 * math.log(1000) + 9) / 10) + 1)
    digits = [0] * pos
    pos -= 1
    digits[pos] = terminator

    while unscaled_value != 0:
        divrem = (unscaled_value // 1000, unscaled_value % 1000)

        pos -= 1
        if pos < 0:
            # Handle rare case when an extra digit is required.
            digits.insert(0, 0)
            pos = 0

        digits[pos] = int(divrem[1]) + digit_adj
        unscaled_value = divrem[0]

    # Now encode digits in proper order, 10 bits per digit. 1024 possible
    # values per 10 bits, and so base 1000 is quite efficient.

    accum = 0
    bits = 0
    for digit in digits:
        accum = (accum << 10) | digit
        bits += 10
        while bits >= 8:
            bits -= 8
            dst[offset] = ((accum >> bits) ^ mask) & 0xff
            offset += 1

    if bits != 0:
        dst[offset] = ((accum << (8 - bits)) ^ mask) & 0xff
        offset += 1

    return offset

def _estimateLength(value_parts):
    # Exactly predicting encoding length is hard, so overestimate to be safe.

    # Calculate number of base-1000 digits, rounded up. Add two more digits to account for
    # the terminator and a rare extra digit.
    # Probably a faster way, given how slow Decimal.as_tuple() is...
    digits = 2 + int(math.ceil(len(value_parts.digits) / 3.0))

    # Calculate the number of bytes required to encode the base-1000 digits, at 10 bits
    # per digit. Add 5 for the maximum header length, and add 7 for round-up behavior when
    # dividing by 8.
    return 5 + (((digits * 10) + 7) >> 3)

def decode_all(data, mask=0, context=None):
    if not isinstance(data, (bytes, bytearray)):
        # TODO or memoryview?
        raise TypeError('data must be bytes or bytearray, got ' + type(data).__name__)

    with localcontext(context or DYNAMODB_CONTEXT) as ctx:
        value, _ = _decode(data, 0, mask, ctx)
        return value

def decode(src, offset, mask=0, context=None):
    if not isinstance(src, (bytes, bytearray)):
        # TODO or memoryview?
        raise TypeError('src must be bytes or bytearray, got ' + type(src).__name__)

    with localcontext(context or DYNAMODB_CONTEXT) as ctx:
        return _decode(src, offset, mask, ctx)

def _decode(src, offset, mask, context):
    orig_offset = offset

    header = six.indexbytes(src, offset) ^ mask

    if header == NULL_BYTE_HIGH or header == NULL_BYTE_LOW:
        return (None, 1)
    elif header == 0x7f or header == 0x80:
        return (context.create_decimal(0), 1)
    elif header == 0x01 or header == 0x7e:
        digit_adj = 999 + 12
        exponent = _readInt32BE(src, offset + 1) ^ mask ^ 0x7fffffff
        if header == 0x7e:
            # Correct negative exponents, as written values are unsigned
            exponent = (exponent & 0x7fffffff) - (exponent & 0x80000000)
        offset += 5
    elif header == 0x81 or header == 0xfe:
        digit_adj = 12
        exponent = _readInt32BE(src, offset + 1) ^ mask ^ 0x80000000
        if header == 0x81:
            # Correct negative exponents, as written values are unsigned
            exponent = (exponent & 0x7fffffff) - (exponent & 0x80000000)
        offset += 5
    else:
        exponent = six.indexbytes(src, offset) ^ mask
        offset += 1
        if exponent >= 0x82:
            digit_adj = 12
            exponent -= 0xc0
        else:
            digit_adj = 999 + 12
            exponent = 0x3f - exponent

    # Significand is base 1000 encoded, 10 bits per digit.
    unscaled_value = None
    precision = 0

    accum = 0
    bits = 0
    last_digit = None

    while True:
        accum = (accum << 8) | (six.indexbytes(src, offset) ^ mask)
        offset += 1
        bits += 8
        if bits >= 10:
            digit = (accum >> (bits - 10)) & 0x3ff

            if digit == 0 or digit == 1023:
                last_digit //= 100
                unscaled_value = last_digit if unscaled_value is None else unscaled_value.fma(10, last_digit)
                precision += 1
                break
            elif digit == 1 or digit == 1022:
                last_digit //= 10
                unscaled_value = last_digit if unscaled_value is None else unscaled_value.fma(100, last_digit)
                precision += 2
                break
            elif digit == 2 or digit == 1021:
                unscaled_value = last_digit if unscaled_value is None else unscaled_value.fma(1000, last_digit)
                precision += 3
                break
            else:
                if unscaled_value is None:
                    unscaled_value = last_digit
                    if unscaled_value is not None:
                        precision += 3
                else:
                    unscaled_value = unscaled_value.fma(1000, last_digit)
                    precision += 3

                bits -= 10
                last_digit = context.create_decimal(digit - digit_adj)

    scale = precision - exponent
    with localcontext() as ctx:
        # Allow extra scale for intermediate values
        ctx.Emax = max(ctx.Emax, -scale)
        scaler = ctx.create_decimal((0, (1,), -scale))
        value = unscaled_value * scaler

    return (value, offset - orig_offset)

def compareUnsigned(a, b):
    if not isinstance(a, (bytes, bytearray)):
        raise TypeError('lhs must be bytes or bytearray')
    if not isinstance(b, (bytes, bytearray)):
        raise TypeError('rhs must be bytes or bytearray')

    return _compareUnsigned(a, 0, len(a), b, 0, len(b))

def _compareUnsigned(a, aoff, alen, b, boff, blen):
    minLen = min(alen, blen)
    for i in range(minLen):
        ab = a[aoff + i]
        bb = b[boff + i]
        if ab != bb:
            return (ab & 0xff) - (bb & 0xff)

    return alen - blen

def _writeInt32BE(buf, value, offset):
    struct.pack_into('>I', buf, offset, value)

def _readInt32BE(buf, offset):
    return struct.unpack_from('>I', buf, offset)[0]



#
#  Copyright 2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License"). You may not
#  use this file except in compliance with the License. A copy of the License
#  is located at
#
#     http://aws.amazon.com/apache2.0/
#
#  or in the "license" file accompanying this file. This file is distributed on
#  an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#  express or implied. See the License for the specific language governing
#  permissions and limitations under the License.
#
from .. import Constants, AttributeValueEncoder





def write_authorizeConnection_1489122155_1(accessKeyId, signature, stringToSign, sessionToken, userAgent, tube):
    tube.write_int(1)
    tube.write_int(1489122155)
    tube.write_string(accessKeyId)
    
    tube.write_string(signature)
    
    tube.write_binary(stringToSign)
    
    if sessionToken is None:
        tube.write_null()
    else:
        tube.write_string(sessionToken)
    
    if userAgent is None:
        tube.write_null()
    else:
        tube.write_string(userAgent)
    

    tube.flush()

def write_batchGetItem_N697851100_1(request, tube):
    tube.write_int(1)
    tube.write_int(-697851100)
    AttributeValueEncoder.encode_batchGetItem_N697851100_1_getItemKeys_direct(tube.enc, request)
    
    
    has_kwargs = (("ReturnConsumedCapacity" in request))
    if has_kwargs:
        tube.enc.append_map_stream_header()
        if 'ReturnConsumedCapacity' in request:
            tube.write_int(Constants.DaxDataRequestParam.ReturnConsumedCapacity)
            tube.write_int(getattr(Constants.ReturnConsumedCapacityValues, request.ReturnConsumedCapacity.upper()))
    
    
        tube.enc.append_break()
    else:
        tube.write_null()

    tube.flush()

def write_batchWriteItem_116217951_1(request, tube):
    tube.write_int(1)
    tube.write_int(116217951)
    AttributeValueEncoder.encode_batchWriteItem_116217951_1_keyValuesByTable_direct(tube.enc, request)
    
    
    has_kwargs = (("ReturnConsumedCapacity" in request) or 
            ("ReturnItemCollectionMetrics" in request))
    if has_kwargs:
        tube.enc.append_map_stream_header()
        if 'ReturnConsumedCapacity' in request:
            tube.write_int(Constants.DaxDataRequestParam.ReturnConsumedCapacity)
            tube.write_int(getattr(Constants.ReturnConsumedCapacityValues, request.ReturnConsumedCapacity.upper()))
    
        if 'ReturnItemCollectionMetrics' in request:
            tube.write_int(Constants.DaxDataRequestParam.ReturnItemCollectionMetrics)
            tube.write_int(getattr(Constants.ReturnItemCollectionMetricsValues, request.ReturnItemCollectionMetrics.upper()))
    
    
        tube.enc.append_break()
    else:
        tube.write_null()

    tube.flush()

def write_defineAttributeList_670678385_1(attributeListId, tube):
    tube.write_int(1)
    tube.write_int(670678385)
    tube.write_int(attributeListId)
    

    tube.flush()

def write_defineAttributeListId_N1230579644_1(attributeNames, tube):
    tube.write_int(1)
    tube.write_int(-1230579644)
    tube.write_array(attributeNames)
    

    tube.flush()

def write_defineKeySchema_N742646399_1(tableName, tube):
    tube.write_int(1)
    tube.write_int(-742646399)
    tube.write_binary(tableName)
    

    tube.flush()

def write_deleteItem_1013539361_1(request, tube):
    tube.write_int(1)
    tube.write_int(1013539361)
    tube.write_binary(request.TableName)
    tube.write_binary(AttributeValueEncoder.encode_key(request.Key, request.key_schema))
    
    has_kwargs = (("ReturnValues" in request) or 
            ("ReturnConsumedCapacity" in request) or 
            ("ReturnItemCollectionMetrics" in request) or 
            ("ConditionExpression" in request) or 
            ("ExpressionAttributeNames" in request) or 
            ("ExpressionAttributeValues" in request))
    if has_kwargs:
        tube.enc.append_map_stream_header()
        if 'ReturnValues' in request:
            tube.write_int(Constants.DaxDataRequestParam.ReturnValues)
            tube.write_int(getattr(Constants.ReturnValuesValues, request.ReturnValues.upper()))
    
        if 'ReturnConsumedCapacity' in request:
            tube.write_int(Constants.DaxDataRequestParam.ReturnConsumedCapacity)
            tube.write_int(getattr(Constants.ReturnConsumedCapacityValues, request.ReturnConsumedCapacity.upper()))
    
        if 'ReturnItemCollectionMetrics' in request:
            tube.write_int(Constants.DaxDataRequestParam.ReturnItemCollectionMetrics)
            tube.write_int(getattr(Constants.ReturnItemCollectionMetricsValues, request.ReturnItemCollectionMetrics.upper()))
    
        if 'ConditionExpression' in request:
            tube.write_int(Constants.DaxDataRequestParam.ConditionExpression)
            # kwargs strings are always written as UTF-8 encoded binary types
            tube.write_binary(request.ConditionExpression.encode('utf8'))
    
        # This operation has expressions, so deal with those together
        AttributeValueEncoder.encode_expressions_direct(tube.enc, request)
    
        tube.enc.append_break()
    else:
        tube.write_null()

    tube.flush()

def write_endpoints_455855874_1(tube):
    tube.write_int(1)
    tube.write_int(455855874)

    tube.flush()

def write_getItem_263244906_1(request, tube):
    tube.write_int(1)
    tube.write_int(263244906)
    tube.write_binary(request.TableName)
    tube.write_binary(AttributeValueEncoder.encode_key(request.Key, request.key_schema))
    
    has_kwargs = (("ConsistentRead" in request) or 
            ("ReturnConsumedCapacity" in request) or 
            ("ProjectionExpression" in request) or 
            ("ExpressionAttributeNames" in request))
    if has_kwargs:
        tube.enc.append_map_stream_header()
        if 'ConsistentRead' in request:
            tube.write_int(Constants.DaxDataRequestParam.ConsistentRead)
            tube.write_boolean(request.ConsistentRead)
    
        if 'ReturnConsumedCapacity' in request:
            tube.write_int(Constants.DaxDataRequestParam.ReturnConsumedCapacity)
            tube.write_int(getattr(Constants.ReturnConsumedCapacityValues, request.ReturnConsumedCapacity.upper()))
    
        # This operation has expressions, so deal with those together
        AttributeValueEncoder.encode_expressions_direct(tube.enc, request)
    
        tube.enc.append_break()
    else:
        tube.write_null()

    tube.flush()

def write_putItem_N2106490455_1(request, tube):
    tube.write_int(1)
    tube.write_int(-2106490455)
    tube.write_binary(request.TableName)
    tube.write_binary(AttributeValueEncoder.encode_key(request.Item, request.key_schema))
    tube.write_binary(AttributeValueEncoder.encode_values(request.Item, request.key_schema, request.attr_names, request.attr_list_id))
    
    has_kwargs = (("ReturnValues" in request) or 
            ("ReturnConsumedCapacity" in request) or 
            ("ReturnItemCollectionMetrics" in request) or 
            ("ConditionExpression" in request) or 
            ("ExpressionAttributeNames" in request) or 
            ("ExpressionAttributeValues" in request))
    if has_kwargs:
        tube.enc.append_map_stream_header()
        if 'ReturnValues' in request:
            tube.write_int(Constants.DaxDataRequestParam.ReturnValues)
            tube.write_int(getattr(Constants.ReturnValuesValues, request.ReturnValues.upper()))
    
        if 'ReturnConsumedCapacity' in request:
            tube.write_int(Constants.DaxDataRequestParam.ReturnConsumedCapacity)
            tube.write_int(getattr(Constants.ReturnConsumedCapacityValues, request.ReturnConsumedCapacity.upper()))
    
        if 'ReturnItemCollectionMetrics' in request:
            tube.write_int(Constants.DaxDataRequestParam.ReturnItemCollectionMetrics)
            tube.write_int(getattr(Constants.ReturnItemCollectionMetricsValues, request.ReturnItemCollectionMetrics.upper()))
    
        if 'ConditionExpression' in request:
            tube.write_int(Constants.DaxDataRequestParam.ConditionExpression)
            # kwargs strings are always written as UTF-8 encoded binary types
            tube.write_binary(request.ConditionExpression.encode('utf8'))
    
        # This operation has expressions, so deal with those together
        AttributeValueEncoder.encode_expressions_direct(tube.enc, request)
    
        tube.enc.append_break()
    else:
        tube.write_null()

    tube.flush()

def write_query_N931250863_1(request, tube):
    tube.write_int(1)
    tube.write_int(-931250863)
    tube.write_binary(request.TableName)
    expressions = AttributeValueEncoder.encode_key_condition_expression_direct(tube.enc, request)
    
    has_kwargs = (("IndexName" in request) or 
            ("Select" in request) or 
            ("Limit" in request) or 
            ("ConsistentRead" in request) or 
            ("ScanIndexForward" in request) or 
            ("ExclusiveStartKey" in request) or 
            ("ReturnConsumedCapacity" in request) or 
            ("ProjectionExpression" in request) or 
            ("FilterExpression" in request) or 
            ("ExpressionAttributeNames" in request) or 
            ("ExpressionAttributeValues" in request))
    if has_kwargs:
        tube.enc.append_map_stream_header()
        if 'IndexName' in request:
            tube.write_int(Constants.DaxDataRequestParam.IndexName)
            # kwargs strings are always written as UTF-8 encoded binary types
            tube.write_binary(request.IndexName.encode('utf8'))
    
        if 'Select' in request:
            tube.write_int(Constants.DaxDataRequestParam.Select)
            tube.write_int(getattr(Constants.SelectValues, request.Select.upper()))
    
        if 'Limit' in request:
            tube.write_int(Constants.DaxDataRequestParam.Limit)
            tube.write_int(request.Limit)
    
        if 'ConsistentRead' in request:
            tube.write_int(Constants.DaxDataRequestParam.ConsistentRead)
            tube.write_int(int(request.ConsistentRead))
    
        if 'ScanIndexForward' in request:
            tube.write_int(Constants.DaxDataRequestParam.ScanIndexForward)
            tube.write_int(int(request.ScanIndexForward))
    
        if 'ExclusiveStartKey' in request:
            tube.write_int(Constants.DaxDataRequestParam.ExclusiveStartKey)
            # No encoder for map so use custom encoder
            AttributeValueEncoder.encode_ExclusiveStartKey_direct(tube.enc, request)
    
        if 'ReturnConsumedCapacity' in request:
            tube.write_int(Constants.DaxDataRequestParam.ReturnConsumedCapacity)
            tube.write_int(getattr(Constants.ReturnConsumedCapacityValues, request.ReturnConsumedCapacity.upper()))
    
        # This operation has expressions, so deal with those together
        # For Query, the expressions are already eval'd for KeyCondExpr
        AttributeValueEncoder.write_expressions_direct(tube.enc, expressions, request)
    
        tube.enc.append_break()
    else:
        tube.write_null()

    tube.flush()

def write_scan_N1875390620_1(request, tube):
    tube.write_int(1)
    tube.write_int(-1875390620)
    tube.write_binary(request.TableName)
    
    has_kwargs = (("IndexName" in request) or 
            ("Limit" in request) or 
            ("Select" in request) or 
            ("ExclusiveStartKey" in request) or 
            ("ReturnConsumedCapacity" in request) or 
            ("TotalSegments" in request) or 
            ("Segment" in request) or 
            ("ProjectionExpression" in request) or 
            ("FilterExpression" in request) or 
            ("ExpressionAttributeNames" in request) or 
            ("ExpressionAttributeValues" in request) or 
            ("ConsistentRead" in request))
    if has_kwargs:
        tube.enc.append_map_stream_header()
        if 'IndexName' in request:
            tube.write_int(Constants.DaxDataRequestParam.IndexName)
            # kwargs strings are always written as UTF-8 encoded binary types
            tube.write_binary(request.IndexName.encode('utf8'))
    
        if 'Limit' in request:
            tube.write_int(Constants.DaxDataRequestParam.Limit)
            tube.write_int(request.Limit)
    
        if 'Select' in request:
            tube.write_int(Constants.DaxDataRequestParam.Select)
            tube.write_int(getattr(Constants.SelectValues, request.Select.upper()))
    
        if 'ExclusiveStartKey' in request:
            tube.write_int(Constants.DaxDataRequestParam.ExclusiveStartKey)
            # No encoder for map so use custom encoder
            AttributeValueEncoder.encode_ExclusiveStartKey_direct(tube.enc, request)
    
        if 'ReturnConsumedCapacity' in request:
            tube.write_int(Constants.DaxDataRequestParam.ReturnConsumedCapacity)
            tube.write_int(getattr(Constants.ReturnConsumedCapacityValues, request.ReturnConsumedCapacity.upper()))
    
        if 'TotalSegments' in request:
            tube.write_int(Constants.DaxDataRequestParam.TotalSegments)
            tube.write_int(request.TotalSegments)
    
        if 'Segment' in request:
            tube.write_int(Constants.DaxDataRequestParam.Segment)
            tube.write_int(request.Segment)
    
        if 'ConsistentRead' in request:
            tube.write_int(Constants.DaxDataRequestParam.ConsistentRead)
            tube.write_int(int(request.ConsistentRead))
    
        # This operation has expressions, so deal with those together
        AttributeValueEncoder.encode_expressions_direct(tube.enc, request)
    
        tube.enc.append_break()
    else:
        tube.write_null()

    tube.flush()

def write_transactGetItems_1866287579_1(request, tube):
    tube.write_int(1)
    tube.write_int(1866287579)
    tube.write_array(request.tableNames)
    
    tube.write_array(request['keys'])
    
    if request.projectionExpressions is None:
        tube.write_null()
    else:
        tube.write_array(request.projectionExpressions)
    
    
    has_kwargs = (("ReturnConsumedCapacity" in request))
    if has_kwargs:
        tube.enc.append_map_stream_header()
        if 'ReturnConsumedCapacity' in request:
            tube.write_int(Constants.DaxDataRequestParam.ReturnConsumedCapacity)
            tube.write_int(getattr(Constants.ReturnConsumedCapacityValues, request.ReturnConsumedCapacity.upper()))
    
    
        tube.enc.append_break()
    else:
        tube.write_null()

    tube.flush()

def write_transactWriteItems_N1160037738_1(request, tube):
    tube.write_int(1)
    tube.write_int(-1160037738)
    tube.write_array(request.operations)
    
    tube.write_array(request.tableNames)
    
    tube.write_array(request['keys'])
    
    if request['values'] is None:
        tube.write_null()
    else:
        tube.write_array(request['values'])
    
    if request.returnValues is None:
        tube.write_null()
    else:
        tube.write_array(request.returnValues)
    
    if request.returnValuesOnConditionCheckFailure is None:
        tube.write_null()
    else:
        tube.write_array(request.returnValuesOnConditionCheckFailure)
    
    if request.conditionExpressions is None:
        tube.write_null()
    else:
        tube.write_array(request.conditionExpressions)
    
    if request.updateExpressions is None:
        tube.write_null()
    else:
        tube.write_array(request.updateExpressions)
    
    
    has_kwargs = (("ReturnConsumedCapacity" in request) or 
            ("ReturnItemCollectionMetrics" in request) or 
            ("ClientRequestToken" in request))
    if has_kwargs:
        tube.enc.append_map_stream_header()
        if 'ReturnConsumedCapacity' in request:
            tube.write_int(Constants.DaxDataRequestParam.ReturnConsumedCapacity)
            tube.write_int(getattr(Constants.ReturnConsumedCapacityValues, request.ReturnConsumedCapacity.upper()))
    
        if 'ReturnItemCollectionMetrics' in request:
            tube.write_int(Constants.DaxDataRequestParam.ReturnItemCollectionMetrics)
            tube.write_int(getattr(Constants.ReturnItemCollectionMetricsValues, request.ReturnItemCollectionMetrics.upper()))
    
        if 'ClientRequestToken' in request:
            tube.write_int(Constants.DaxDataRequestParam.ClientRequestToken)
            # ClientRequestToken is written as plain string unlike other string kwargs
            tube.write_string(request.ClientRequestToken)
    
    
        tube.enc.append_break()
    else:
        tube.write_null()

    tube.flush()

def write_updateItem_1425579023_1(request, tube):
    tube.write_int(1)
    tube.write_int(1425579023)
    tube.write_binary(request.TableName)
    tube.write_binary(AttributeValueEncoder.encode_key(request.Key, request.key_schema))
    
    has_kwargs = (("ReturnValues" in request) or 
            ("ReturnConsumedCapacity" in request) or 
            ("ReturnItemCollectionMetrics" in request) or 
            ("UpdateExpression" in request) or 
            ("ConditionExpression" in request) or 
            ("ExpressionAttributeNames" in request) or 
            ("ExpressionAttributeValues" in request))
    if has_kwargs:
        tube.enc.append_map_stream_header()
        if 'ReturnValues' in request:
            tube.write_int(Constants.DaxDataRequestParam.ReturnValues)
            tube.write_int(getattr(Constants.ReturnValuesValues, request.ReturnValues.upper()))
    
        if 'ReturnConsumedCapacity' in request:
            tube.write_int(Constants.DaxDataRequestParam.ReturnConsumedCapacity)
            tube.write_int(getattr(Constants.ReturnConsumedCapacityValues, request.ReturnConsumedCapacity.upper()))
    
        if 'ReturnItemCollectionMetrics' in request:
            tube.write_int(Constants.DaxDataRequestParam.ReturnItemCollectionMetrics)
            tube.write_int(getattr(Constants.ReturnItemCollectionMetricsValues, request.ReturnItemCollectionMetrics.upper()))
    
        if 'ConditionExpression' in request:
            tube.write_int(Constants.DaxDataRequestParam.ConditionExpression)
            # kwargs strings are always written as UTF-8 encoded binary types
            tube.write_binary(request.ConditionExpression.encode('utf8'))
    
        # This operation has expressions, so deal with those together
        AttributeValueEncoder.encode_expressions_direct(tube.enc, request)
    
        tube.enc.append_break()
    else:
        tube.write_null()

    tube.flush()




#!/usr/bin/env python
#
# Copyright 2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import codecs
import os.path
import re
import sys

from setuptools import setup
from setuptools.command.test import test as TestCommand

PY2 = sys.version_info[0] == 2

here = os.path.abspath(os.path.dirname(__file__))

class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass into py.test")]

    def __init__(self, *args, **kwargs):
        super(PyTest, self).__init__(*args, **kwargs)
        self.pytest_args = None
        self.test_suite = None

    def initialize_options(self):
        TestCommand.initialize_options(self)
        try:
            from multiprocessing import cpu_count
            self.pytest_args = ['-n', str(cpu_count()), '--boxed']
        except (ImportError, NotImplementedError):
            self.pytest_args = ['-n', '1', '--boxed']

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest

        errno = pytest.main(self.pytest_args)
        sys.exit(errno)

def read(*parts):
    return codecs.open(os.path.join(here, *parts), 'r').read()

def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

def read_requirements(filename):
    with open(filename) as requirements:
        return [package.strip() for package in requirements]

packages = [
    'amazondax',
    'amazondax.generated',
    # Both versions must be included since we don't know where it will end up running
    'amazondax.grammar',
    'amazondax.grammar2'
]


install_requires = read_requirements("requirements.txt")
tests_require = read_requirements('requirements-dev.txt')

setup_params = dict(
    name='amazon-dax-client',
    version=find_version("amazondax", "__init__.py"),
    author='Amazon Web Services',
    packages=packages,
    url='https://aws.amazon.com/dynamodb/dax/',
    scripts=[],
    license='Apache License 2.0',
    description='Amazon DAX Client for Python',
    long_description=open('README.rst').read(),
    install_requires=install_requires,
    tests_require=tests_require,
    cmdclass={'test': PyTest},
    python_requires='>=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*',
    classifiers=(
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ),
)

setup(**setup_params)



# Generated from grammar/DynamoDbGrammar.g4 by ANTLR 4.7.2
# pylint: disable=W
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .DynamoDbGrammarParser import DynamoDbGrammarParser
else:
    from DynamoDbGrammarParser import DynamoDbGrammarParser

# This class defines a complete listener for a parse tree produced by DynamoDbGrammarParser.
class DynamoDbGrammarListener(ParseTreeListener):

    # Enter a parse tree produced by DynamoDbGrammarParser#projection_.
    def enterProjection_(self, ctx:DynamoDbGrammarParser.Projection_Context):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#projection_.
    def exitProjection_(self, ctx:DynamoDbGrammarParser.Projection_Context):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#projection.
    def enterProjection(self, ctx:DynamoDbGrammarParser.ProjectionContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#projection.
    def exitProjection(self, ctx:DynamoDbGrammarParser.ProjectionContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#condition_.
    def enterCondition_(self, ctx:DynamoDbGrammarParser.Condition_Context):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#condition_.
    def exitCondition_(self, ctx:DynamoDbGrammarParser.Condition_Context):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#Or.
    def enterOr(self, ctx:DynamoDbGrammarParser.OrContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#Or.
    def exitOr(self, ctx:DynamoDbGrammarParser.OrContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#Negation.
    def enterNegation(self, ctx:DynamoDbGrammarParser.NegationContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#Negation.
    def exitNegation(self, ctx:DynamoDbGrammarParser.NegationContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#In.
    def enterIn(self, ctx:DynamoDbGrammarParser.InContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#In.
    def exitIn(self, ctx:DynamoDbGrammarParser.InContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#And.
    def enterAnd(self, ctx:DynamoDbGrammarParser.AndContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#And.
    def exitAnd(self, ctx:DynamoDbGrammarParser.AndContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#Between.
    def enterBetween(self, ctx:DynamoDbGrammarParser.BetweenContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#Between.
    def exitBetween(self, ctx:DynamoDbGrammarParser.BetweenContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#FunctionCondition.
    def enterFunctionCondition(self, ctx:DynamoDbGrammarParser.FunctionConditionContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#FunctionCondition.
    def exitFunctionCondition(self, ctx:DynamoDbGrammarParser.FunctionConditionContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#Comparator.
    def enterComparator(self, ctx:DynamoDbGrammarParser.ComparatorContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#Comparator.
    def exitComparator(self, ctx:DynamoDbGrammarParser.ComparatorContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#ConditionGrouping.
    def enterConditionGrouping(self, ctx:DynamoDbGrammarParser.ConditionGroupingContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#ConditionGrouping.
    def exitConditionGrouping(self, ctx:DynamoDbGrammarParser.ConditionGroupingContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#comparator_symbol.
    def enterComparator_symbol(self, ctx:DynamoDbGrammarParser.Comparator_symbolContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#comparator_symbol.
    def exitComparator_symbol(self, ctx:DynamoDbGrammarParser.Comparator_symbolContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#update_.
    def enterUpdate_(self, ctx:DynamoDbGrammarParser.Update_Context):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#update_.
    def exitUpdate_(self, ctx:DynamoDbGrammarParser.Update_Context):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#update.
    def enterUpdate(self, ctx:DynamoDbGrammarParser.UpdateContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#update.
    def exitUpdate(self, ctx:DynamoDbGrammarParser.UpdateContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#set_section.
    def enterSet_section(self, ctx:DynamoDbGrammarParser.Set_sectionContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#set_section.
    def exitSet_section(self, ctx:DynamoDbGrammarParser.Set_sectionContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#set_action.
    def enterSet_action(self, ctx:DynamoDbGrammarParser.Set_actionContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#set_action.
    def exitSet_action(self, ctx:DynamoDbGrammarParser.Set_actionContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#add_section.
    def enterAdd_section(self, ctx:DynamoDbGrammarParser.Add_sectionContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#add_section.
    def exitAdd_section(self, ctx:DynamoDbGrammarParser.Add_sectionContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#add_action.
    def enterAdd_action(self, ctx:DynamoDbGrammarParser.Add_actionContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#add_action.
    def exitAdd_action(self, ctx:DynamoDbGrammarParser.Add_actionContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#delete_section.
    def enterDelete_section(self, ctx:DynamoDbGrammarParser.Delete_sectionContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#delete_section.
    def exitDelete_section(self, ctx:DynamoDbGrammarParser.Delete_sectionContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#delete_action.
    def enterDelete_action(self, ctx:DynamoDbGrammarParser.Delete_actionContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#delete_action.
    def exitDelete_action(self, ctx:DynamoDbGrammarParser.Delete_actionContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#remove_section.
    def enterRemove_section(self, ctx:DynamoDbGrammarParser.Remove_sectionContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#remove_section.
    def exitRemove_section(self, ctx:DynamoDbGrammarParser.Remove_sectionContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#remove_action.
    def enterRemove_action(self, ctx:DynamoDbGrammarParser.Remove_actionContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#remove_action.
    def exitRemove_action(self, ctx:DynamoDbGrammarParser.Remove_actionContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#OperandValue.
    def enterOperandValue(self, ctx:DynamoDbGrammarParser.OperandValueContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#OperandValue.
    def exitOperandValue(self, ctx:DynamoDbGrammarParser.OperandValueContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#ArithmeticValue.
    def enterArithmeticValue(self, ctx:DynamoDbGrammarParser.ArithmeticValueContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#ArithmeticValue.
    def exitArithmeticValue(self, ctx:DynamoDbGrammarParser.ArithmeticValueContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#PlusMinus.
    def enterPlusMinus(self, ctx:DynamoDbGrammarParser.PlusMinusContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#PlusMinus.
    def exitPlusMinus(self, ctx:DynamoDbGrammarParser.PlusMinusContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#ArithmeticParens.
    def enterArithmeticParens(self, ctx:DynamoDbGrammarParser.ArithmeticParensContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#ArithmeticParens.
    def exitArithmeticParens(self, ctx:DynamoDbGrammarParser.ArithmeticParensContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#PathOperand.
    def enterPathOperand(self, ctx:DynamoDbGrammarParser.PathOperandContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#PathOperand.
    def exitPathOperand(self, ctx:DynamoDbGrammarParser.PathOperandContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#LiteralOperand.
    def enterLiteralOperand(self, ctx:DynamoDbGrammarParser.LiteralOperandContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#LiteralOperand.
    def exitLiteralOperand(self, ctx:DynamoDbGrammarParser.LiteralOperandContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#FunctionOperand.
    def enterFunctionOperand(self, ctx:DynamoDbGrammarParser.FunctionOperandContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#FunctionOperand.
    def exitFunctionOperand(self, ctx:DynamoDbGrammarParser.FunctionOperandContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#ParenOperand.
    def enterParenOperand(self, ctx:DynamoDbGrammarParser.ParenOperandContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#ParenOperand.
    def exitParenOperand(self, ctx:DynamoDbGrammarParser.ParenOperandContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#FunctionCall.
    def enterFunctionCall(self, ctx:DynamoDbGrammarParser.FunctionCallContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#FunctionCall.
    def exitFunctionCall(self, ctx:DynamoDbGrammarParser.FunctionCallContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#path.
    def enterPath(self, ctx:DynamoDbGrammarParser.PathContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#path.
    def exitPath(self, ctx:DynamoDbGrammarParser.PathContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#id_.
    def enterId_(self, ctx:DynamoDbGrammarParser.Id_Context):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#id_.
    def exitId_(self, ctx:DynamoDbGrammarParser.Id_Context):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#MapAccess.
    def enterMapAccess(self, ctx:DynamoDbGrammarParser.MapAccessContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#MapAccess.
    def exitMapAccess(self, ctx:DynamoDbGrammarParser.MapAccessContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#ListAccess.
    def enterListAccess(self, ctx:DynamoDbGrammarParser.ListAccessContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#ListAccess.
    def exitListAccess(self, ctx:DynamoDbGrammarParser.ListAccessContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#LiteralSub.
    def enterLiteralSub(self, ctx:DynamoDbGrammarParser.LiteralSubContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#LiteralSub.
    def exitLiteralSub(self, ctx:DynamoDbGrammarParser.LiteralSubContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#expression_attr_names_sub.
    def enterExpression_attr_names_sub(self, ctx:DynamoDbGrammarParser.Expression_attr_names_subContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#expression_attr_names_sub.
    def exitExpression_attr_names_sub(self, ctx:DynamoDbGrammarParser.Expression_attr_names_subContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#expression_attr_values_sub.
    def enterExpression_attr_values_sub(self, ctx:DynamoDbGrammarParser.Expression_attr_values_subContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#expression_attr_values_sub.
    def exitExpression_attr_values_sub(self, ctx:DynamoDbGrammarParser.Expression_attr_values_subContext):
        pass


    # Enter a parse tree produced by DynamoDbGrammarParser#unknown.
    def enterUnknown(self, ctx:DynamoDbGrammarParser.UnknownContext):
        pass

    # Exit a parse tree produced by DynamoDbGrammarParser#unknown.
    def exitUnknown(self, ctx:DynamoDbGrammarParser.UnknownContext):
        pass



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import threading
import random
import time

import logging
logger = logging.getLogger(__name__)

# TODO There should be one of these per-process, and it should handle multiple tasks. That way there's only one thread.
class PeriodicTask(threading.Thread):
    def __init__(self, period, func, args=None, kwargs=None, jitter=None, clock=time.time):
        super(PeriodicTask, self).__init__(name='PeriodicTask-' + getattr(func, '__name__', ''))
        self.daemon = True # Ensure tasks do not block shutdown

        self.period = period
        self.func = func
        self.args = args if args is not None else []
        self.kwargs = kwargs if kwargs is not None else {}
        self.jitter = jitter
        self.tick = threading.Event()
        self.starttime = clock()
        self._clock = clock

    @property
    def cancelled(self):
        return self.tick.is_set()

    def cancel(self):
        # This flags the event (and thus task) as cancelled
        self.tick.set()

    def run(self):
        while not self.tick.is_set():
            # Add random jitter if requested
            jit = random.uniform(-self.jitter, self.jitter) if self.jitter else 0.0 # pylint: disable=invalid-unary-operand-type

            # Wait for the given interval
            self.tick.wait(self.period + jit)

            if not self.tick.is_set():
                # If it hasn't been cancelled, run the function
                self.func(*self.args, **self.kwargs)

def periodic_task(func, period, jitter=None):
    ''' period -> seconds '''
    task_man = PeriodicTask(period, func, jitter=jitter)
    task_man.start()
    return task_man



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import time
import threading
import socket
import binascii
import datetime

from random import SystemRandom

from . import SigV4Gen, __version__ as pkg_version
from .CborEncoder import CborEncoder
from .DaxCborDecoder import DaxCborDecoder
from .Constants import STRING_TYPES, BINARY_TYPES, INTEGRAL_TYPES, SEQ_TYPES

import logging
logger = logging.getLogger(__name__)

class Tube(object):
    MAGIC_STRING = b'gJ7yne5G'
    USER_AGENT_STRING = u'UserAgent'
    DEFAULT_USER_AGENT = u'DaxPythonClient-' + pkg_version
    WINDOW_SCALAR = 0.1
    DAX_ADDR = u'https://dax.amazonaws.com'
    DEFAULT_FLUSH_SIZE = 4096
    MIN_READ_SIZE = 4096

    def __init__(self, sock, version, cred_provider, region, user_agent=None, user_agent_extra=None, clock=time.time):
        self._socket = sock
        self._version = version
        self._cred_provider = cred_provider
        self._region = region
        self._access_key_id = None
        self._auth_exp = 0
        self._auth_ttl_millis = 5 * 60 * 1000
        self._pool_window = (self._auth_ttl_millis / 2)
        self._tube_window = (self._auth_ttl_millis * Tube.WINDOW_SCALAR)
        self._last_pool_auth = 0
        self.clock = clock

        self._user_agent = user_agent or Tube.DEFAULT_USER_AGENT
        if user_agent_extra:
            self._user_agent += ' ' + user_agent_extra

        self.enc = CborEncoder()
        self.dec = DaxCborDecoder(self._more)

        self._connect(self._version)

    def _connect(self, session):
        # Send the initialization info to the server
        self.enc.append_raw(Tube.MAGIC_STRING)
        self.enc.append_int(0)     # No layering

        if session is None:
            self.enc.append_null()
        else:
            self.enc.append_binary(session.as_bytes())

        self.enc.append_map_header(1)
        self.enc.append_string(Tube.USER_AGENT_STRING)
        self.enc.append_string(self._user_agent)

        self.enc.append_int(0)     # No types

        self.flush()

    def _more(self, buf, n):
        # TODO This may be more efficient with a memoryview on the buffer
        # Although the rest of the buffer handling needs to be rewritten as well

        # Attempt to read at least MIN_READ_SIZE to avoid lots of 1 byte reads.
        data = self._socket.recv(max(n, Tube.MIN_READ_SIZE))
        buf += data

        return len(data)

    def close(self):
        ''' Close the underlying socket. '''
        if self._socket:
            self._socket.close()
            self._socket = None

    def reauth(self):
        ''' Resend the authentication request (if necessary) before the next API call. '''
        # Run everything in millis to be consistent with other clients
        now = self.clock() * 1000

        if self._auth_exp - now <= self._tube_window or now - self._last_pool_auth >= self._pool_window:
            creds = self._cred_provider()
            self._checkAndUpdateAccessKeyId(creds.access_key)
            self._last_pool_auth = now
            self._auth_exp = now + self._auth_ttl_millis

            self._auth(creds, now/1000)

    def write(self, b):
        ''' Write raw bytes to the destination. '''
        self.enc.append_raw(b)
        self._maybe_flush()

    def write_object(self, o):
        ''' Write an object of unknown type to the destination.

        Only Python primitives (and their subclasses) are supported, not arbitrary objects.
        '''
        if o is None:
            self.write_null()
        elif isinstance(o, bool):
            self.write_boolean(o)
        elif isinstance(o, INTEGRAL_TYPES):
            self.write_int(o)
        elif isinstance(o, STRING_TYPES):
            self.write_string(o)
        elif isinstance(o, BINARY_TYPES):
            self.write_binary(o)
        elif isinstance(o, SEQ_TYPES):
            self.write_array(o)
        elif isinstance(o, dict):
            self.write_map(o)
        else:
            raise ValueError('Unknown type: ' + type(o).__name__)

        # Flush is performed by the type-specific function

    def write_int(self, i):
        ''' Write a CBOR int record (Python int) to the destination. '''
        self.enc.append_int(i)
        self._maybe_flush()

    def write_binary(self, b):
        ''' Write a CBOR bytes record (Python bytes or bytearray) to the destination. '''
        self.enc.append_binary(b)
        self._maybe_flush()

    def write_string(self, s):
        ''' Write a CBOR text record (Python str) to the destination. '''
        self.enc.append_string(s)
        self._maybe_flush()

    def write_array(self, a):
        ''' Write a CBOR array (Python seq/list/tuple) to the destination. '''
        self.enc.append_array_header(len(a))
        for item in a:
            self.write_object(item)
        self._maybe_flush()

    def write_map(self, m):
        ''' Write a CBOR array (Python dict) to the destination. '''
        self.enc.append_map_header(len(m))
        for key, value in m.items():
            self.write_object(key)
            self.write_object(value)
        self._maybe_flush()

    def write_boolean(self, b):
        ''' Write a CBOR boolean of the appropriate value.'''
        self.enc.append_boolean(b)
        self._maybe_flush()

    def write_null(self):
        ''' Write a CBOR NULL record to the destination. '''
        self.enc.append_null()
        self._maybe_flush()

    def peek(self):
        ''' Peek at the next CBOR record. '''
        return self.dec.peek()

    def skip(self):
        ''' Skip the next CBOR record. '''
        return self.dec.skip()

    def read_int(self):
        ''' Read the next int. '''
        return self.dec.decode_int()

    def read_string(self):
        ''' Read the next int. '''
        return self.dec.decode_string()

    def read_object(self):
        ''' Read the next object. '''
        return self.dec.decode_object()

    def read_map(self):
        ''' Read a map. '''
        return self.dec.decode_map()

    def read_array(self):
        ''' Read an array. '''
        return self.dec.decode_array()

    def try_read_null(self):
        ''' Check for CBOR NULL.

        If the next token is a CBOR NULL, consume it and return True. Otherwise,
        leave it and return False.
        '''
        return self.dec.try_decode_null()

    def _maybe_flush(self):
        ''' Flush some data if the buffer is larger than the current flush size. '''
        if len(self.enc._buffer) > Tube.DEFAULT_FLUSH_SIZE:
            self._flush_some()

    def _flush_some(self):
        ''' Flush as much as possible in one call. '''
        n = self._socket.send(self.enc._buffer)
        self.enc._drain(n)

    def flush(self):
        ''' Completely flush the write buffer. '''
        self._socket.sendall(self.enc._buffer)
        self.enc.reset()

    def clear(self):
        self.enc.reset()

    def _auth(self, creds, now):
        ''' Send an authentication RPC prior to the next request. '''
        logger.debug('Sending authentication request')
        dtnow = datetime.datetime.utcfromtimestamp(now)
        sig = SigV4Gen.generate_signature(creds, Tube.DAX_ADDR, self._region, b'', dtnow)

        self.enc.append_int(1)
        self.enc.append_int(1489122155) # authorizeConnection method ID
        self.enc.append_string(creds.access_key)
        self.enc.append_string(sig.signature)
        self.enc.append_binary(sig.string_to_sign)

        token = getattr(sig, 'token')
        if token:
            self.enc.append_string(token)
        else:
            self.enc.append_null()

        if self._user_agent:
            self.enc.append_string(self._user_agent)
        else:
            self.enc.append_null()

    def _checkAndUpdateAccessKeyId(self, akid):
        if not akid:
            raise ValueError('AWSCredentialsProvider provided null AWSAccessKeyId')

        if akid != self._access_key_id:
            self._access_key_id = akid
            return True
        else:
            return False

class TubePool(object):
    def __init__(self):
        self._head_tube = None
        self._last_active_tube = None
        self._session_version = _SessionVersion()
        self._connection = 0
        self._total = 0

        self._lock = threading.RLock()
        self._avail = threading.Condition(self._lock)

        # TODO Start a Timer job to reap idle tubes

    def get(self):
        with self._lock:
            while True:
                tube = self._head_tube
                self._total += 1
                if tube is not None:
                    self._head_tube = tube._next_tube
                    if self._last_active_tube is tube:
                        self._last_active_tube = self._head_tube
                    tube._next_tube = None
                    return _TubeManager(self, tube)
                else:
                    self._connection += 1
                    self._lock.release()
                    try:
                        # Connect outside of the lock
                        # TODO Ideally this is an asynchronous process
                        # Otherwise connect() can block until available, when a
                        # tube may get recycled sooner
                        # The Java client uses a future; perhaps we could do that
                        # here as well
                        new_tube = self._alloc()
                    finally:
                        self._lock.acquire()

                    self.recycle(new_tube)

    def recycle(self, tube):
        if tube is None:
            return

        with self._lock:
            if tube._version == self._session_version:
                tube._next_tube = self._head_tube
                self._head_tube = tube
                self._avail.notify()
                return

        tube.close()

    def reset(self, tube):
        if tube is None:
            return

        tube.close()

        with self._lock:
            if tube._version != self._session_version:
                return

            self._version_bump()
            tube = self._head_tube
            self._head_tube = self._last_active_tube = None
            self._avail.notify_all()

        self._close_all(tube)

    def close(self):
        with self._lock:
            self._version_bump()
            head = self._head_tube
            self._head_tube = self._last_active_tube = None
        self._close_all(head)

    def reap_idle_tubes(self):
        to_reap = None
        with self._lock:
            # Select last active tube
            last_active_tube = self._last_active_tube
            # If it's not null, break chain after this tube
            if last_active_tube is not None:
                to_reap = last_active_tube._next_tube
                last_active_tube._next_tube = None

            self._last_active_tube = self._head_tube

        # Close tubes downstream of the to_reap one
        self._close_all(to_reap)
        self._connection = 0

    def _close_all(self, head_tube):
        reap_count = 0
        tube = head_tube
        while tube is not None:
            reap_count += 1
            tube.close()

            _next = tube._next_tube
            tube._next_tube = None
            tube = _next

        return reap_count

    def _version_bump(self):
        self._session_version = _SessionVersion()

    def _alloc(self):
        raise NotImplementedError('abstract method')

class SocketTubePool(TubePool):
    DEFAULT_CONNECT_TIMEOUT = 1 # seconds
    DEFAULT_READ_TIMEOUT = 60   # seconds

    def __init__(self, hostname, port, cred_provider, region, user_agent=None, user_agent_extra=None,
                 connect_timeout=None, read_timeout=None):
        super(SocketTubePool, self).__init__()
        self.hostname = hostname
        self.port = port
        self.cred_provider = cred_provider
        self.region = region
        self.user_agent = user_agent
        self.user_agent_extra = user_agent_extra
        self.connect_timeout = connect_timeout if connect_timeout is not None else SocketTubePool.DEFAULT_CONNECT_TIMEOUT
        self.read_timeout = read_timeout if read_timeout is not None else SocketTubePool.DEFAULT_READ_TIMEOUT

    def _alloc(self):
        logger.debug('SocketTubePool: New connection to %s:%s', self.hostname, self.port)
        sock = socket.create_connection((self.hostname, self.port), timeout=self.connect_timeout)

        # set socket options NODELAY, timeouts, keepalives
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        sock.settimeout(self.read_timeout)

        return Tube(sock,
            self._session_version,
            self.cred_provider,
            self.region,
            self.user_agent,
            self.user_agent_extra)

class _TubeManager(object):
    def __init__(self, pool, tube):
        self._tube = tube
        self._pool = pool

    def __enter__(self):
        return self._tube

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            # Normal exit case
            self._pool.recycle(self._tube)
        else:
            # An exception was thrown
            self._pool.reset(self._tube)

_version_rand = SystemRandom()
class _SessionVersion(object):
    def __init__(self):
        self.session = _version_rand.getrandbits(128)

    def __eq__(self, other):
        return self.session == other.session

    def as_bytes(self):
        # Convert the integer to bytes, regardless of integer size
        hval = '%x' % self.session
        bval = binascii.unhexlify('0' + hval if len(hval) % 2 != 0 else hval)

        return bval



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

TAG_DDB_STRING_SET = 3321
TAG_DDB_NUMBER_SET = 3322
TAG_DDB_BINARY_SET = 3323
TAG_DDB_DOCUMENT_PATH_ORDINAL = 3324

class DdbSet(object):
    def __init__(self, set_type, values):
        self.values = values
        self.type = set_type

    def __repr__(self):
        return 'DdbSet({}, {})'.format(self.type, self.values)

class DocumentPathOrdinal(object):
    def __init__(self, ordinal):
        self.ordinal = ordinal

